from enum import Enum
from typing import Dict, Any


class ResponseStatus(Enum):
    Success = 0
    AlreadyRegistered = 1
    NotRegistered = 2


class CallbackRegistrationResponse:
    def __init__(self, callback_id, response_status=None):
        self._callback_id = callback_id
        self._response_status = response_status

    @property
    def callback_id(self):
        return self._callback_id

    @property
    def response_status(self):
        return self._response_status


class CallbackFunction:
    def __init__(self, callback, *args):
        self._callback = callback
        self._args = args

    def call(self) -> Any:
        return self._callback(*self._args)


class CallbackRegistry:
    def __init__(self):
        self._registered_callbacks = {}

    def register_change_callback(self, callback_id, change_callback, *args):
        if callback_id in self._registered_callbacks:
            return CallbackRegistrationResponse(callback_id, ResponseStatus.AlreadyRegistered)
        self._registered_callbacks[callback_id] = CallbackFunction(change_callback, *args)
        return CallbackRegistrationResponse(callback_id, ResponseStatus.Success)

    def unregister_change_callback(self, callback_id):
        if callback_id not in self._registered_callbacks:
            return CallbackRegistrationResponse(callback_id, ResponseStatus.NotRegistered)
        self._registered_callbacks[callback_id] = None
        return CallbackRegistrationResponse(callback_id, ResponseStatus.Success)

    def call(self) -> Dict[str, Any]:
        return {callback_id: callback.call() for callback_id, callback in self._registered_callbacks.items()}
