from asset_discovery.models import RefreshReason
from asset_discovery.models.discovery_type import DiscoveryType
from asset_discovery.notification.callback_registry import CallbackRegistry
from asset_discovery.util.asset_serializer import AssetSerializer
from asset_discovery.repository.file_json_provider import FileJsonProvider
from asset_discovery.repository.json_data_store import JsonDataStore
from asset_discovery.plugins.json_imports.settings_provider import SettingsProvider
from asset_discovery.util.named_mutex import NamedMutex
from asset_discovery.util.named_mutex_exception import NamedMutexTimeoutException
import logging

import os
if os.name == 'nt':
    from asset_discovery.file_monitor.file_monitor import FileMonitor
else:
    from asset_discovery.file_monitor.linux_file_monitor import LinuxFileMonitor as FileMonitor

logger = logging.getLogger()


class AssetPlugin:
    _JSON_MUTEX_NAME = "asset_discovery.plugins.json"

    def __init__(self, json_provider=None):
        settings_provider = SettingsProvider()
        store_directory_path = settings_provider.get_store_directory_path()
        if json_provider is None:
            json_provider = FileJsonProvider(store_directory_path)
        self._json_asset_data_store = JsonDataStore(json_provider, AssetSerializer())
        self._callback_registry = CallbackRegistry()
        self._monitor = FileMonitor(
            store_directory_path,
            file_extension=".json",
            callback_registry=self._callback_registry)

    def __del__(self):
        self._monitor.stop_monitoring()

    @staticmethod
    def _log_error(e):
        logger.error(e)

    def list_assets(self, refresh_reason=RefreshReason.USER_REQUEST, **kwargs):
        assets = self._json_asset_data_store.list()
        assets = self._set_discovery_type_manual(assets)
        return assets

    def add_asset(self, asset, **kwargs):
        if asset.discovery_type is not None:
            asset.discovery_type = None
        try:
            with NamedMutex(AssetPlugin._JSON_MUTEX_NAME):
                self._json_asset_data_store.add(asset)
        except NamedMutexTimeoutException as e:
            self._log_error(e)

    def remove_asset(self, asset_identification, **kwargs):
        try:
            with NamedMutex(AssetPlugin._JSON_MUTEX_NAME):
                self._json_asset_data_store.remove(asset_identification)
        except NamedMutexTimeoutException as e:
            self._log_error(e)

    def update_asset(self, asset_update, **kwargs):
        partial_update = kwargs.get("partial_update", False)
        try:
            with NamedMutex(AssetPlugin._JSON_MUTEX_NAME):
                self._json_asset_data_store.update(asset_update, partial_update)
        except NamedMutexTimeoutException as e:
            self._log_error(e)

    def register_change_callback(self, callback_id, callback):
        plugin_id = __name__.split('.')[-1]
        self._callback_registry.register_change_callback(callback_id, callback, plugin_id)
        self._start_monitoring()

    def _start_monitoring(self):
        return self._monitor.start_monitoring()

    @staticmethod
    def _set_discovery_type_manual(assets):
        for asset in assets:
            asset.discovery_type = DiscoveryType.MANUAL
        return assets
