from abc import ABC, abstractmethod
from datetime import datetime
from typing import Optional

from asset_discovery.models import SensorCollection, SensorModel, SensorProperties, SensorHealthState
from asset_discovery.plugins.sysapi_imports.repository.models import PropertyBag, AttributeSensorHealthState


class AbstractSensorSerializer(ABC):
    SENSOR_HEALTH_STATE_VALUES = {
        AttributeSensorHealthState.Unknown: SensorHealthState.UNKNOWN,
        AttributeSensorHealthState.Normal: SensorHealthState.NORMAL,
        AttributeSensorHealthState.ExceedsUserDefinedThreshold: SensorHealthState.EXCEEDS_USER_DEFINED_THRESHOLD,
        AttributeSensorHealthState.OutOfSpec: SensorHealthState.OUT_OF_SPEC,
        AttributeSensorHealthState.RiskOfDamage: SensorHealthState.RISK_OF_DAMAGE
    }

    def __init__(self, sensor_type, base_name, numeric_attribute_tags):
        """
        :param sensor_type: The type of the sensor that will be parsed.
        :type: SensorType

        :param base_name: The default name base for the sensors which have no name; the
            name will be computed as ``base_nameN`` where N is the index of the sensor

        :param numeric_attribute_tags: Defines the numeric attribute tags corresponding to
            the property tags of the sensor types. This makes sense as the 3 sensor types have
            the same attribute types (ex name, reading etc) but are stored at different keys.
        :type: BaseSensorAttributes
        """
        self.sensor_type = sensor_type
        self.numeric_attribute_tags = numeric_attribute_tags
        self.base_name = base_name

    def serialize(self, bag: PropertyBag, timestamp: datetime) -> Optional[SensorCollection]:
        count = bag.get_property_or_default(self.numeric_attribute_tags.count, 0)
        if count == 0:
            return None
        sensors = []
        for i in range(count):
            sensor = self.__serialize_base(i, bag)
            sensor = self._on_sensor_parsing(i, bag, sensor)
            sensor.sensor_type = self.sensor_type
            sensors.append(sensor)
        return SensorCollection(sensors, timestamp)

    def __serialize_base(self, i, bag):
        sensor = SensorModel()
        sensor.name = self.__get_name_from_bag(i, bag)
        sensor.reading = bag.get_property_or_default(self.numeric_attribute_tags.reading0 + i, None)
        state = bag.get_property_or_default(self.numeric_attribute_tags.health_state0 + i, None)
        sensor.state = self.SENSOR_HEALTH_STATE_VALUES.get(state, None)
        alarm = bag.get_property_or_default(self.numeric_attribute_tags.alarm0 + i, None)
        sensor.alarm = self.SENSOR_HEALTH_STATE_VALUES.get(alarm, None)

        upper_critical = bag.get_property_or_default(self.numeric_attribute_tags.upper_critical0 + i, None)
        upper_damage_risk = bag.get_property_or_default(self.numeric_attribute_tags.upper_damage_risk0 + i, None)
        upper_user = bag.get_property_or_default(self.numeric_attribute_tags.upper_user0 + i, None)
        accuracy0 = bag.get_property_or_default(self.numeric_attribute_tags.accuracy0 + i, None)
        polling_interval0 = bag.get_property_or_default(self.numeric_attribute_tags.polling_interval0 + i, None)

        properties = {}

        if upper_critical is not None:
            properties[SensorProperties.UpperCritical] = upper_critical
        if upper_damage_risk is not None:
            properties[SensorProperties.UpperRisk] = upper_damage_risk
        if upper_user is not None:
            properties[SensorProperties.UpperUser] = upper_user
        if accuracy0 is not None:
            properties[SensorProperties.Accuracy] = accuracy0
        if polling_interval0 is not None:
            properties[SensorProperties.Interval] = polling_interval0

        sensor.properties = properties
        return sensor

    def __get_name_from_bag(self, i, bag):
        default_name = "{0}{1}".format(self.base_name, i)
        return bag.get_property_or_default(self.numeric_attribute_tags.name0 + i, default_name)

    @abstractmethod
    def _on_sensor_parsing(self, i, bag, base_sensor):
        pass
