import os
import subprocess
import threading

try:
    import tornado as tornado
    import tornado.ioloop as tornado_ioloop
except ImportError:
    import salt.ext.tornado as tornado
    import salt.ext.tornado.ioloop as tornado_ioloop

from asset_discovery.models.operation_result import OperationResult
from asset_discovery.models.operation_status import OperationStatus

try:
    import pythoncom
    import wmi
except ImportError:
    # We are on Linux, this is not an issue
    pass

from asset_discovery.plugins.sysapi_imports.repository.bag_provider import BagProvider
from asset_discovery.util.named_mutex import NamedMutex
from asset_discovery.plugins.sysapi_imports.repository.sms_ipc_client import SMSIPCClient
from asset_discovery.plugins.sysapi_imports.repository.models import PropertyBag


class Com:  # This has been recreated from salt.salt.utils.winapi.Com()
    def __init__(self):
        self.need_com_init = not self._is_main_thread()

    def _is_main_thread(self):
        return threading.current_thread().name == "MainThread"

    def __enter__(self):
        if self.need_com_init:
            # log.debug("Initializing COM library")
            pythoncom.CoInitialize()

    def __exit__(self, exc_type, exc_value, traceback):
        if self.need_com_init:
            # log.debug("Uninitializing COM library")
            pythoncom.CoUninitialize()


class IPCBagProvider(BagProvider):
    def __init__(self, connect_timeout=5):
        # TODO: All configs, constants will be extracted
        self._service_name = 'niminionagent' if self._is_windows() else ''
        self._daemon_path = '/etc/init.d/niminionagent' if not self._is_windows() else ''
        self._operation_timed_out_code = -2147220448
        self._service_not_found_code = -2147220557
        self._mutex_name = 'salt.modules.nisysapi'
        self._connect_timeout = connect_timeout

    def get_all(self):
        if not self._agent_enabled():
            return []
        else:
            response = self._run_request("get_props", [], {})
            resources = response.get("resources", [])
            return self._create_bags(resources)

    def _agent_enabled(self):
        '''
        Return True if the National Instruments Minion Agent is running.
        False otherwise.
        '''
        if self._is_windows():
            try:
                with Com():
                    wmi_c = wmi.WMI()
                    matched_services = wmi_c.Win32_Service(name=self._service_name)
                    if not matched_services:
                        return False
                    return matched_services[0].State == 'Running'
            except Exception as exc:  # pylint: disable=broad-except
                # if not WMI_FAILURE_LOGGED:
                #     WMI_FAILURE_LOGGED = True
                #     log.error(
                #         'Exception occurred when using WMI Win32_Service: %s',
                #         exc, exc_info=True
                #     )
                # On WMI Error, treat as if the Minion Agent is not running.
                return False
        else:  # NOTE: in Salt here would be an explicit check for Linux; we don't need that
            if not os.path.isfile(self._daemon_path):
                return False
            return_code = subprocess.call([self._daemon_path, 'status'])
            return return_code == 0

    @staticmethod
    def _is_windows():
        return os.name == "nt"

    def _run_request(self, function, args, kwargs):
        request = {
            'fun': "nisysapi.{0}".format(function),
            'args': args,
            'kwargs': kwargs,
            'client': 'ni_asset',
            'jid': 'SysAPI Plugin Job',  # TODO: We should make this not required in MinionAgent
            'id': 'Current System'  # TODO: We should make this not required in MinionAgent
        }
        return self._query_agent(request, self._connect_timeout)

    def _query_agent(self, request, connect_timeout=5):
        '''
        Send a job request to the NI Minion Agent
        and receive a response with the job return info.
        '''
        mutex = NamedMutex(self._mutex_name)
        with mutex:
            io_loop = tornado_ioloop.IOLoop()
            client = None
            try:
                uri = 7890
                client = SMSIPCClient(uri, io_loop=io_loop)
                try:
                    io_loop.run_sync(
                        lambda: client.connect(timeout=connect_timeout))
                except Exception:  # pylint: disable=broad-except
                    # log.error('Connection failed to: %s', uri)
                    return {'status_code': self._service_not_found_code}

                try:
                    # log.info('About to send request to: %s', uri)
                    response = io_loop.run_sync(
                        lambda: client.query(request))
                except Exception as exc:  # pylint: disable=broad-except
                    # log.error('Request to: %s failed. Exception = %s', uri, exc)
                    return {'status_code': self._operation_timed_out_code}
            finally:
                if client is not None:
                    client.close()
                io_loop.close()

        # log.info('Response received!')
        return response

    @staticmethod
    def _create_bags(resources):
        bags = [IPCBagProvider._create_bag(resource) for resource in resources]
        return [bag for bag in bags if not bag.is_empty()]

    @staticmethod
    def _create_bag(resource):
        resource_uri = resource.get("resource_uri", "")
        bag = PropertyBag(resource_uri)
        properties = resource.get("properties", [])
        for prop in properties:
            bag.add_property(prop["tag"], prop["value"])
        return bag

    def get_unsafe(self, requested_bags):
        resource_tags = [{"resource_uri": req.resource_uri, "tags": req.properties} for req in requested_bags]
        request = {"resource_tags": resource_tags}
        response = self._run_request('populate_props', [], request)
        resources = response.get("resources", [])
        return self._create_bags(resources)

    def create(self, request_kwargs):
        if not self._agent_enabled():
            status = OperationStatus(successful=False, code=self._service_not_found_code, description="Service not found")
            return OperationResult(status=status)
        response = self._run_request("create_resource", [], request_kwargs)
        status_code = response["status_code"]
        status = OperationStatus(successful=True, code=status_code)
        if status_code < 0:
            status = OperationStatus(successful=False, code=status_code)
        return OperationResult(status=status)

    def update(self, asset_update_request):
        if not self._agent_enabled():
            return {'status_code': self._service_not_found_code}
        resource_json = asset_update_request.to_dict()
        asset_update_request = {'resources': [resource_json]}
        return self._run_request('set_props', [], asset_update_request)

    def remove(self, resource_uri: str, mode: int):
        if not self._agent_enabled():
            return {'status_code': self._service_not_found_code}
        job_kwargs = {
            'resource_uri': resource_uri,
            'mode': mode
        }
        return self._run_request('delete_resource', [], job_kwargs)

    def self_test(self, bag, mode):
        if not self._agent_enabled():
            return {'status_code': self._service_not_found_code}
        job_kwargs = {
            'resource_uri': bag.resource_uri,
            'mode': mode
        }
        return self._run_request('self_test', [], job_kwargs)

    def self_calibrate(self, bag):
        if not self._agent_enabled():
            return {'status_code': self._service_not_found_code}
        job_kwargs = {'resource_uri': bag.resource_uri}
        return self._run_request('self_cal', [], job_kwargs)

    def self_reset(self, bag, mode):
        if not self._agent_enabled():
            return {'status_code': self._service_not_found_code}
        resource_uri = bag.resource_uri
        reset_request = {'resource_uri': resource_uri, 'mode': mode}
        return self._run_request('reset_device', [], reset_request)

    def any_properties_changed(self):
        if not self._agent_enabled():
            return {'status_code': self._service_not_found_code}
        response = self._run_request('props_changed', [], None)
        if isinstance(response, dict):
            return response.get('status_code', -1) >= 0 and response.get('props_changed', False) is True
