from typing import Dict

from asset_discovery.models import Asset, OperationResult, OperationStatus
from asset_discovery.plugins.sysapi_imports.models.add_device_type import AddDeviceType
from asset_discovery.plugins.sysapi_imports.request_helpers.add.add_network_daqmx_request_builder import \
    AddNetworkDaqmxRequestBuilder
from asset_discovery.plugins.sysapi_imports.request_helpers.add.add_network_visa_request_builder import \
    AddNetworkVisaRequestBuilder
from asset_discovery.plugins.sysapi_imports.request_helpers.add.add_request_builder_base import AddRequestBuilderBase


class AddRequestBuilder(AddRequestBuilderBase):

    def __init__(self):
        self._helpers = self._get_add_request_helpers()

    @classmethod
    def _get_add_request_helpers(cls) -> Dict[str, AddRequestBuilderBase]:
        return {
            AddDeviceType.NETWORK_DAQMX: AddNetworkDaqmxRequestBuilder(),
            AddDeviceType.NETWORK_VISA: AddNetworkVisaRequestBuilder()
        }

    def build_from(self, asset: Asset, **kwargs) -> OperationResult:
        device_type = kwargs.get("device_type", None)
        validation_result = self._validate(device_type, asset)
        if not validation_result.status.succeeded:
            return validation_result
        return self._helpers[device_type].build_from(asset, **kwargs)

    def _validate(self, device_type: str, asset: Asset) -> OperationResult:
        if not device_type:
            status = OperationStatus(
                successful=False,
                code=-1,
                description="Missing argument \"device_type\"."
            )
            return OperationResult(status=status)

        if not self._supports_device_type(device_type):
            status = OperationStatus(
                successful=False,
                code=-1,
                description="Invalid value for \"device_type\""
            )
            return OperationResult(status)

        if (not asset.ip_addresses or len(asset.ip_addresses) == 0) and not asset.hostname:
            status = OperationStatus(
                successful=False,
                code=-1,
                description="Need to specify at least one of: ip address and hostname."
            )
            return OperationResult(status=status)

        status = OperationStatus(successful=True, code=0)
        return OperationResult(status)

    def _supports_device_type(self, device_type: str):
        return device_type in self._helpers

